/** \file MiniscopeDeviceInterface.cpp
*/

/* Copyright (C) 2008-2009 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//---------------------------------------------------------------------------

#pragma hdrstop

#include "MiniscopeDeviceInterface.h"
#include "MiniscopeDevice.h"
#include "Log.h"

//---------------------------------------------------------------------------

#pragma package(smart_init)

std::vector<DllInfo> DeviceInterface::dlls;
std::set<class DeviceInterface*> DeviceInterface::instances;


void DeviceInterface::EnumerateDlls(AnsiString dir)
{
	LOG(E_LOG_TRACE, "Enumerating device dlls...");
	DeviceInterface::dlls.clear();
	WIN32_FIND_DATA file;
	if (dir[dir.Length()] != '\\')
		dir += "\\";
	AnsiString asSrchPath = dir + "*.dll";
	HANDLE hFind = FindFirstFile(asSrchPath.c_str(), &file);
	int hasfiles = (hFind != INVALID_HANDLE_VALUE);
	struct DllInfo dllinfo;

	while (hasfiles)
	{
		AnsiString filename = dir + file.cFileName;
		if(VerifyDll(filename, &dllinfo) == E_OK)
		{
			AddDll(dllinfo);
		} else
			LOG(E_LOG_TRACE, "Library %s not loaded", filename.c_str());
		hasfiles = FindNextFile(hFind, &file);
	}
	FindClose(hFind);
}

enum DeviceInterface::E_LIB_STATUS DeviceInterface::VerifyDll(AnsiString filename, struct DllInfo* const dllinfo)
{
#if 0
	// Suppressing "Cannot Find ThisLibrary.dll" or "Cannot Find SomeOtherRequired.dll" Error Box
	UINT uOldErrorMode = SetErrorMode(SEM_NOOPENFILEERRORBOX | SEM_FAILCRITICALERRORS);
	HINSTANCE hInstance = LoadLibrary(filename.c_str());
	SetErrorMode(uOldErrorMode);

	if ((int)hInstance == 2)
	{
		//MessageBox(NULL, "MYDLL.DLL does not exist", "ERROR", MB_OK);
		return false;
	}
	else if ((int)hInstance <= HINSTANCE_ERROR)
	{
		//MessageBox(NULL, "Other problems loading MYDLL.DLL, "ERROR", MB_OK);
		return false;
	}
#else
	// Windows error message when loading may actually be usefull when resolving additional dll dependencies
	HINSTANCE hInstance = LoadLibrary(filename.c_str());
	if (!hInstance)
	{
		LOG(E_LOG_TRACE, "Failed to load %s library", filename.c_str());
		return E_LOADFAILED;
	}
#endif

	pfGetMiniscopeInterfaceDescription gmid =
		(pfGetMiniscopeInterfaceDescription)GetProcAddress(hInstance, "GetMiniscopeInterfaceDescription");
	if (!gmid)
	{
		LOG(E_LOG_TRACE, "Library %s DOES NOT look like a compatible plugin", filename.c_str());
		FreeLibrary(hInstance);
		return E_NOTVALID;
	}
	struct S_DEVICE_DLL_INTERFACE interf_version;
	gmid(&interf_version);
	if (interf_version.majorVersion != DLL_INTERFACE_MAJOR_VERSION ||
		interf_version.minorVersion != DLL_INTERFACE_MINOR_VERSION)
	{
		LOG(E_LOG_TRACE, "Library %s has interface v.%d.%d instead of required %d.%d",
			filename.c_str(),
			interf_version.majorVersion, interf_version.minorVersion,
			DLL_INTERFACE_MAJOR_VERSION);
		FreeLibrary(hInstance);
		return E_INCOMPATIBLE_VER;
	}

	FreeLibrary(hInstance);
	dllinfo->name = filename;
	dllinfo->ver.majorVersion = interf_version.majorVersion;
	dllinfo->ver.minorVersion = interf_version.minorVersion;
	return E_OK;
}

void DeviceInterface::AddDll(const struct DllInfo &dllinfo)
{
	LOG(E_LOG_TRACE, "Adding %s", dllinfo.name.c_str());
	DeviceInterface::dlls.push_back(dllinfo);
}

void __stdcall DeviceInterface::OnLog(void *cookie, char *szText)
{
#if 0
	class DeviceInterface *dev;
	dev = reinterpret_cast<class DeviceInterface*>(cookie);
	if (instances.find(dev) == instances.end())
	{
		LOG(E_LOG_TRACE, "OnLog called with unknown cookie %p. No matching object.", dev);
		return;
	}
#endif
	LOG(E_LOG_TRACE, "%s", szText);
}

void __stdcall DeviceInterface::OnConnect(void *cookie, int state, char *szMsgText)
{
	class DeviceInterface *dev;
	dev = reinterpret_cast<class DeviceInterface*>(cookie);
	if (instances.find(dev) == instances.end())
	{
		//LOG(E_LOG_TRACE, "OnConnect called with unknown cookie %p. No matching object.", dev);
		return;
	}
	dev->connInfo.state = static_cast<enum E_CONNECTION_STATE>(state);
	dev->connInfo.msg = szMsgText;
	if (dev->callbackConnect)
		dev->callbackConnect(state, szMsgText);
}

void __stdcall DeviceInterface::OnData(void *cookie, void *buffer, unsigned int buffersize)
{
	class DeviceInterface *dev;
	dev = reinterpret_cast<class DeviceInterface*>(cookie);
	if (instances.find(dev) == instances.end())
	{
		//LOG(E_LOG_TRACE, "OnData called with unknown cookie %p. No matching object.", dev);
		return;
	}
	if (dev->callbackData)
		dev->callbackData((short*)buffer, buffersize);
}

void __stdcall DeviceInterface::OnTrigger(void *cookie)
{
	class DeviceInterface *dev;
	dev = reinterpret_cast<class DeviceInterface*>(cookie);
	if (instances.find(dev) == instances.end())
	{
		//LOG(E_LOG_TRACE, "OnData called with unknown cookie %p. No matching object.", dev);
		return;
	}
	if (dev->callbackTrigger)
		dev->callbackTrigger();
}




DeviceInterface::DeviceInterface(AnsiString asDllName):
	hInstance(NULL),
	filename(asDllName),
	capabilities(NULL),
	callbackData(NULL),
	callbackConnect(NULL),
	dllGetMiniscopeInterfaceDescription(NULL),
	dllGetMiniscopeSettings(NULL),
	dllSaveMiniscopeSettings(NULL),
	dllSetCallbacks(NULL),
	dllShowSettings(NULL),
	dllGetDeviceCapabilities(NULL), dllSetSensitivity(NULL),
	dllSetSamplingPeriod(NULL), dllSetTriggerSource(NULL),
	dllSetTriggerLevel(NULL),
	dllSetAcqMode(NULL),
	dllSetCouplingType(NULL), dllSetBufferSize(NULL),
	dllConnect(NULL), dllDisconnect(NULL),
	dllSetOffset(NULL), dllAcquireStart(NULL),
	dllAcquireStop(NULL), dllManualTrigger(NULL),
	dllStoreCalibration(NULL), dllGetCalibration(NULL)
	//connInfo.state( {DEVICE_DISCONNECTED, "Not connected"} ),
{
	LOG(E_LOG_TRACE, "Creating object using %s", asDllName.c_str());
	connInfo.state = DEVICE_DISCONNECTED;
	connInfo.msg = "Not connected";
	instances.insert(this);
}

DeviceInterface::~DeviceInterface()
{
	instances.erase(this);
	if (hInstance)
		FreeLibrary(hInstance);
}

int DeviceInterface::Load(void)
{
	hInstance = LoadLibrary(filename.c_str());
	if (!hInstance)
		return 1;
	dllSetCallbacks = (pfSetCallbacks)GetProcAddress(hInstance, "SetCallbacks");
	dllShowSettings = (pfShowSettings)GetProcAddress(hInstance, "ShowSettings");
	dllGetDeviceCapabilities = (pfGetDeviceCapabilities)GetProcAddress(hInstance, "GetDeviceCapabilities");

	dllGetMiniscopeSettings = (pfGetMiniscopeSettings)GetProcAddress(hInstance, "GetMiniscopeSettings");
	dllSaveMiniscopeSettings = (pfSaveMiniscopeSettings)GetProcAddress(hInstance, "SaveMiniscopeSettings");

	dllSetSensitivity = (pfSetSensitivity)GetProcAddress(hInstance, "SetSensitivity");
	dllSetSamplingPeriod = (pfSetSamplingPeriod)GetProcAddress(hInstance, "SetSamplingPeriod");
	dllSetTriggerSource = (pfSetTriggerSource)GetProcAddress(hInstance, "SetTriggerSource");
    dllSetAcqMode = (pfSetAcqMode)GetProcAddress(hInstance, "SetAcqMode");
	dllSetTriggerSource = (pfSetTriggerSource)GetProcAddress(hInstance, "SetTriggerSource");
    dllSetTriggerLevel = (pfSetTriggerLevel)GetProcAddress(hInstance, "SetTriggerLevel");
	dllSetCouplingType = (pfSetCouplingType)GetProcAddress(hInstance, "SetCouplingType");
	dllSetBufferSize = (pfSetBufferSize)GetProcAddress(hInstance, "SetBufferSize");
	dllSetOffset = (pfSetOffset)GetProcAddress(hInstance, "SetOffset");
	dllConnect = (pfConnect)GetProcAddress(hInstance, "Connect");
	dllDisconnect = (pfDisconnect)GetProcAddress(hInstance, "Disconnect");
	dllAcquireStart = (pfAcquireStart)GetProcAddress(hInstance, "AcquireStart");
	dllAcquireStop = (pfAcquireStop)GetProcAddress(hInstance, "AcquireStop");
	dllManualTrigger = (pfManualTrigger)GetProcAddress(hInstance, "ManualTrigger");
	dllStoreCalibration = (pfStoreCalibration)GetProcAddress(hInstance, "StoreCalibration");
	dllGetCalibration = (pfGetCalibration)GetProcAddress(hInstance, "GetCalibration");

	if ((dllSetCallbacks && dllShowSettings &&
		dllGetDeviceCapabilities && dllConnect &&
		dllDisconnect && dllSetAcqMode && dllAcquireStart && dllAcquireStop) == 0)
	{
		LOG(E_LOG_TRACE, "Dll load failed. Some of the required functions are missing");
		return 2;
	}

	dllSetCallbacks(this, &OnLog, &OnConnect, &OnData, &OnTrigger);
	dllGetDeviceCapabilities(&this->capabilities);
	return 0;
}


